@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 # Copyright, 2017, by Samuel G. D. Williams. <http://www.codeotaku.com>
 # 
 # Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -23,15 +25,24 @@ require_relative 'notification'
 module Async
 	# A queue which allows items to be processed in order.
 	class Queue < Notification
-		def initialize
-			super
+		def initialize(parent: nil)
+			super()
 			
 			@items = []
+			@parent = parent
 		end
 		
 		attr :items
 		
-		def enqueue item
+		def size
+			@items.size
+		end
+
+		def empty?
+			@items.empty?
+		end
+		
+		def enqueue(item)
 			@items.push(item)
 			
 			self.signal unless self.empty?
@@ -47,21 +58,26 @@ module Async
 			@items.shift
 		end
 		
-		def async(&block)
-			parent = Task.current
-			
+		def async(parent: (@parent or Task.current), &block)
 			while item = self.dequeue
 				parent.async(item, &block)
 			end
 		end
+		
+		def each
+			while item = self.dequeue
+				yield item
+			end
+		end
 	end
 	
 	class LimitedQueue < Queue
-		def initialize(limit = 1)
-			super()
+		def initialize(limit = 1, **options)
+			super(**options)
 			
 			@limit = limit
-			@full = Async::Queue.new
+			
+			@full = Notification.new
 		end
 		
 		attr :limit
@@ -72,8 +88,8 @@ module Async
 		end
 		
 		def enqueue item
-			if limited?
-				@full.dequeue
+			while limited?
+				@full.wait
 			end
 			
 			super
@@ -82,7 +98,7 @@ module Async
 		def dequeue
 			item = super
 			
-			@full.enqueue(nil) unless @full.empty?
+			@full.signal
 			
 			return item
 		end
