@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 # Copyright, 2017, by Samuel G. D. Williams. <http://www.codeotaku.com>
 # 
 # Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -21,6 +23,7 @@
 require_relative 'logger'
 require_relative 'task'
 require_relative 'wrapper'
+require_relative 'scheduler'
 
 require 'nio'
 require 'timers'
@@ -42,16 +45,14 @@ module Async
 		# - When invoked at the top level, will create and run a reactor, and invoke
 		# the block as an asynchronous task. Will block until the reactor finishes
 		# running.
-		def self.run(*args, **options, &block)
+		def self.run(*arguments, **options, &block)
 			if current = Task.current?
-				reactor = current.reactor
-				
-				return reactor.async(*args, **options, &block)
+				return current.async(*arguments, **options, &block)
 			else
-				reactor = self.new(**options)
+				reactor = self.new
 				
 				begin
-					return reactor.run(*args, &block)
+					return reactor.run(*arguments, **options, &block)
 				ensure
 					reactor.close
 				end
@@ -80,27 +81,67 @@ module Async
 			@ready = []
 			@running = []
 			
-			@stopped = true
+			if Scheduler.supported?
+				@scheduler = Scheduler.new(self)
+			else
+				@scheduler = nil
+			end
+			
+			@interrupted = false
+			@guard = Mutex.new
+			@blocked = 0
+			@unblocked = []
 		end
 		
-		def logger
-			@logger ||= Console.logger
+		attr :scheduler
+		attr :logger
+		
+		# @reentrant Not thread safe.
+		def block(blocker, timeout)
+			fiber = Fiber.current
+			
+			if timeout
+				timer = @timers.after(timeout) do
+					if fiber.alive?
+						fiber.resume(false)
+					end
+				end
+			end
+			
+			begin
+				@blocked += 1
+				Task.yield
+			ensure
+				@blocked -= 1
+			end
+		ensure
+			timer&.cancel
+		end
+		
+		# @reentrant Thread safe.
+		def unblock(blocker, fiber)
+			@guard.synchronize do
+				@unblocked << fiber
+				@selector.wakeup
+			end
+		end
+		
+		def fiber(&block)
+			if @scheduler
+				Fiber.new(blocking: false, &block)
+			else
+				Fiber.new(&block)
+			end
 		end
 		
 		def to_s
-			"\#<#{self.description} (#{@stopped ? 'stopped' : 'running'})>"
+			"\#<#{self.description} #{@children&.size || 0} children (#{stopped? ? 'stopped' : 'running'})>"
 		end
 		
-		# @attr stopped [Boolean]
-		attr :stopped
-		
 		def stopped?
-			@stopped
+			@children.nil?
 		end
 		
-		# TODO Remove these in next major release. They are too confusing to use correctly.
-		def_delegators :@timers, :every, :after
-		
 		# Start an asynchronous task within the specified reactor. The task will be
 		# executed until the first blocking call, at which point it will yield and
 		# and this method will return.
@@ -109,7 +150,7 @@ module Async
 		#
 		# @yield [Task] Executed within the task.
 		# @return [Task] The task that was scheduled into the reactor.
-		def async(*args, **options, &block)
+		def async(*arguments, **options, &block)
 			task = Task.new(self, **options, &block)
 			
 			# I want to take a moment to explain the logic of this.
@@ -119,9 +160,9 @@ module Async
 			# - Fail at the point of the method call where possible.
 			# - Execute determinstically where possible.
 			# - Avoid scheduler overhead if no blocking operation is performed.
-			task.run(*args)
+			task.run(*arguments)
 			
-			# logger.debug "Initial execution of task #{fiber} complete (#{result} -> #{fiber.alive?})..."
+			# Console.logger.debug "Initial execution of task #{fiber} complete (#{result} -> #{fiber.alive?})..."
 			return task
 		end
 		
@@ -132,12 +173,13 @@ module Async
 			return monitor
 		end
 		
-		# Stop the reactor at the earliest convenience. Can be called from a different thread safely.
-		# @return [void]
-		def stop
-			unless @stopped
-				@stopped = true
-				@selector.wakeup
+		# Interrupt the reactor at the earliest convenience. Can be called from a different thread safely.
+		def interrupt
+			@guard.synchronize do
+				unless @interrupted
+					@interrupted = true
+					@selector.wakeup
+				end
 			end
 		end
 		
@@ -151,79 +193,120 @@ module Async
 		def yield(fiber = Fiber.current)
 			@ready << fiber
 			
-			Fiber.yield
+			Task.yield
 		end
 		
 		def finished?
-			# I'm not sure if checking `@running.empty?` is really required.
-			super && @ready.empty? && @running.empty?
+			# TODO I'm not sure if checking `@running.empty?` is really required.
+			super && @ready.empty? && @running.empty? && @blocked.zero?
 		end
 		
-		# Run the reactor until either all tasks complete or {#stop} is invoked.
-		# Proxies arguments to {#async} immediately before entering the loop.
-		def run(*args, &block)
+		# Run one iteration of the event loop.
+		# @param timeout [Float | nil] the maximum timeout, or if nil, indefinite.
+		# @return [Boolean] whether there is more work to do.
+		def run_once(timeout = nil)
+			# Console.logger.debug(self) {"@ready = #{@ready} @running = #{@running}"}
+			
+			if @ready.any?
+				# running used to correctly answer on `finished?`, and to reuse Array object.
+				@running, @ready = @ready, @running
+				
+				@running.each do |fiber|
+					fiber.resume if fiber.alive?
+				end
+				
+				@running.clear
+			end
+			
+			if @unblocked.any?
+				unblocked = Array.new
+				
+				@guard.synchronize do
+					unblocked, @unblocked = @unblocked, unblocked
+				end
+				
+				while fiber = unblocked.pop
+					fiber.resume if fiber.alive?
+				end
+			end
+			
+			if @ready.empty?
+				interval = @timers.wait_interval
+			else
+				# if there are tasks ready to execute, don't sleep:
+				interval = 0
+			end
+			
+			# If we are finished, we stop the task tree and exit:
+			if self.finished?
+				return false
+			end
+			
+			# If there is no interval to wait (thus no timers), and no tasks, we could be done:
+			if interval.nil?
+				# Allow the user to specify a maximum interval if we would otherwise be sleeping indefinitely:
+				interval = timeout
+			elsif interval < 0
+				# We have timers ready to fire, don't sleep in the selctor:
+				interval = 0
+			elsif timeout and interval > timeout
+				interval = timeout
+			end
+			
+			# Console.logger.info(self) {"Selecting with #{@children&.size} children with interval = #{interval ? interval.round(2) : 'infinite'}..."}
+			if monitors = @selector.select(interval)
+				monitors.each do |monitor|
+					monitor.value.resume
+				end
+			end
+			
+			@timers.fire
+			
+			# We check and clear the interrupted flag here:
+			if @interrupted
+				@guard.synchronize do
+					@interrupted = false
+				end
+				
+				return false
+			end
+			
+			# The reactor still has work to do:
+			return true
+		end
+		
+		# Run the reactor until all tasks are finished. Proxies arguments to {#async} immediately before entering the loop, if a block is provided.
+		def run(*arguments, **options, &block)
 			raise RuntimeError, 'Reactor has been closed' if @selector.nil?
 			
-			@stopped = false
+			@scheduler&.set!
 			
-			initial_task = self.async(*args, &block) if block_given?
+			initial_task = self.async(*arguments, **options, &block) if block_given?
 			
-			until @stopped
-				logger.debug(self) {"@ready = #{@ready} @running = #{@running}"}
-				
-				if @ready.any?
-					# running used to correctly answer on `finished?`, and to reuse Array object.
-					@running, @ready = @ready, @running
-					
-					@running.each do |fiber|
-						fiber.resume if fiber.alive?
-					end
-					
-					@running.clear
-				end
-				
-				if @ready.empty?
-					interval = @timers.wait_interval
-				else
-					# if there are tasks ready to execute, don't sleep:
-					interval = 0
-				end
-				
-				# If there is no interval to wait (thus no timers), and no tasks, we could be done:
-				if interval.nil?
-					if self.finished?
-						# If there is nothing to do, then finish:
-						return initial_task
-					end
-				elsif interval < 0
-					# We have timers ready to fire, don't sleep in the selctor:
-					interval = 0
-				end
-				
-				logger.debug(self) {"Selecting with #{@children&.size} children with interval = #{interval ? interval.round(2) : 'infinite'}..."}
-				if monitors = @selector.select(interval)
-					monitors.each do |monitor|
-						monitor.value.resume
-					end
-				end
-				
-				@timers.fire
+			while self.run_once
+				# Round and round we go!
 			end
 			
 			return initial_task
 		ensure
-			logger.debug(self) {"Exiting run-loop because #{$! ? $! : 'finished'}."}
-			
-			@stopped = true
+			@scheduler&.clear!
+			Console.logger.debug(self) {"Exiting run-loop because #{$! ? $! : 'finished'}."}
 		end
-	
+		
+		def stop(later = true)
+			@children&.each do |child|
+				# We don't want this process to propagate `Async::Stop` exceptions, so we schedule tasks to stop later.
+				child.stop(later)
+			end
+		end
+		
 		# Stop each of the children tasks and close the selector.
 		# 
 		# @return [void]
 		def close
-			@children&.each(&:stop)
+			# This is a critical step. Because tasks could be stored as instance variables, and since the reactor is (probably) going out of scope, we need to ensure they are stopped. Otherwise, the tasks will belong to a reactor that will never run again and are not stopped.
+			self.stop(false)
 			
-			# TODO Should we also clear all timers?
 			@selector.close
 			@selector = nil
 		end
@@ -233,13 +316,13 @@ module Async
 		def closed?
 			@selector.nil?
 		end
-	
+		
 		# Put the calling fiber to sleep for a given ammount of time.
 		# @param duration [Numeric] The time in seconds, to sleep for.
 		def sleep(duration)
 			fiber = Fiber.current
 			
-			timer = self.after(duration) do
+			timer = @timers.after(duration) do
 				if fiber.alive?
 					fiber.resume
 				end
@@ -256,10 +339,10 @@ module Async
 		def with_timeout(timeout, exception = TimeoutError)
 			fiber = Fiber.current
 			
-			timer = self.after(timeout) do
+			timer = @timers.after(timeout) do
 				if fiber.alive?
 					error = exception.new("execution expired")
-					fiber.resume error
+					fiber.resume(error)
 				end
 			end
 			
@@ -267,12 +350,5 @@ module Async
 		ensure
 			timer.cancel if timer
 		end
-		
-		# TODO remove
-		def timeout(*args, &block)
-			warn "#{self.class}\#timeout(...) is deprecated, use #{self.class}\#with_timeout(...) instead."
-			
-			with_timeout(*args, &block)
-		end
 	end
 end
