@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 # Copyright, 2017, by Samuel G. D. Williams. <http://www.codeotaku.com>
 # 
 # Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -70,31 +72,36 @@ module Async
 		# Create a new task.
 		# @param reactor [Async::Reactor] the reactor this task will run within.
 		# @param parent [Async::Task] the parent task.
-		def initialize(reactor, parent = Task.current?, logger: nil, &block)
-			super(parent || reactor)
+		def initialize(reactor, parent = Task.current?, logger: nil, finished: nil, **options, &block)
+			super(parent || reactor, **options)
 			
 			@reactor = reactor
 			
 			@status = :initialized
 			@result = nil
-			@finished = nil
+			@finished = finished
 			
-			@logger = logger
+			@logger = logger || @parent.logger
 			
 			@fiber = make_fiber(&block)
 		end
 		
+		attr :logger
+		
+		if Fiber.current.respond_to?(:backtrace)
+			def backtrace(*arguments)
+				@fiber&.backtrace(*arguments)
+			end
+		end
+		
 		def to_s
 			"\#<#{self.description} (#{@status})>"
 		end
 		
-		def logger
-			@logger ||= @parent&.logger
-		end
-		
 		# @attr ios [Reactor] The reactor the task was created within.
 		attr :reactor
-		def_delegators :@reactor, :with_timeout, :timeout, :sleep
+		
+		def_delegators :@reactor, :with_timeout, :sleep
 		
 		# Yield back to the reactor and allow other fibers to execute.
 		def yield
@@ -112,19 +119,20 @@ module Async
 		attr :status
 		
 		# Begin the execution of the task.
-		def run(*args)
+		def run(*arguments)
 			if @status == :initialized
 				@status = :running
-				@fiber.resume(*args)
+				
+				@fiber.resume(*arguments)
 			else
 				raise RuntimeError, "Task already running!"
 			end
 		end
 		
-		def async(*args, **options, &block)
+		def async(*arguments, **options, &block)
 			task = Task.new(@reactor, self, **options, &block)
 			
-			task.run(*args)
+			task.run(*arguments)
 			
 			return task
 		end
@@ -223,7 +231,7 @@ module Async
 		private
 		
 		# This is a very tricky aspect of tasks to get right. I've modelled it after `Thread` but it's slightly different in that the exception can propagate back up through the reactor. If the user writes code which raises an exception, that exception should always be visible, i.e. cause a failure. If it's not visible, such code fails silently and can be very difficult to debug.
-		# As an explcit choice, the user can start a task which doesn't propagate exceptions. This only applies to `StandardError` and derived tasks. This allows tasks to internally capture their error state which is raised when invoking `Task#result` similar to how `Thread#join` works. This mode makes `Async::Task` behave more like a promise, and you would need to ensure that someone calls `Task#result` otherwise you might miss important errors.
+		# As an explcit choice, the user can start a task which doesn't propagate exceptions. This only applies to `StandardError` and derived tasks. This allows tasks to internally capture their error state which is raised when invoking `Task#result` similar to how `Thread#join` works. This mode makes {ruby Async::Task} behave more like a promise, and you would need to ensure that someone calls `Task#result` otherwise you might miss important errors.
 		def fail!(exception = nil, propagate = true)
 			@status = :failed
 			@result = exception
@@ -232,9 +240,9 @@ module Async
 				raise
 			elsif @finished.nil?
 				# If no one has called wait, we log this as an error:
-				logger.error(self) {$!}
+				Console.logger.error(self) {$!}
 			else
-				logger.debug(self) {$!}
+				Console.logger.debug(self) {$!}
 			end
 		end
 		
@@ -248,13 +256,13 @@ module Async
 		end
 		
 		def make_fiber(&block)
-			Fiber.new do |*args|
+			Fiber.new do |*arguments|
 				set!
 				
 				begin
-					@result = yield(self, *args)
+					@result = yield(self, *arguments)
 					@status = :complete
-					# logger.debug(self) {"Task was completed with #{@children.size} children!"}
+					# Console.logger.debug(self) {"Task was completed with #{@children.size} children!"}
 				rescue Stop
 					stop!
 				rescue StandardError => error
@@ -262,7 +270,7 @@ module Async
 				rescue Exception => exception
 					fail!(exception, true)
 				ensure
-					# logger.debug(self) {"Task ensure $!=#{$!} with #{@children.size} children!"}
+					# Console.logger.debug(self) {"Task ensure $!=#{$!} with #{@children.size} children!"}
 					finish!
 				end
 			end
@@ -286,6 +294,7 @@ module Async
 		def set!
 			# This is actually fiber-local:
 			Thread.current[:async_task] = self
+			Console.logger = @logger if @logger
 		end
 	end
 end
